var Q = require('q');
var oujRedis = require('./oujRedis.js');
var TableHelper = require('./TableHelper.js');
var r2m_configs = require('../../conf/r2m_config.inc.js');
var util = require('util');

/**
 * 基本类，提供增删改查
 * @param {string} dbKey 
 */
function Redis2MySql(tableName, dbKey, cacheKey) {
    this.tableName = tableName;
    this.dbKey = dbKey || 'default';
    this.cacheKey = cacheKey || 'default';
    this.cacheInfo = r2m_configs[this.dbKey][tableName];
    if (this.cacheInfo) {
        throw new Error("redis没配置table name:" + tableName, CODE_REDIS_ERROR);
    }

    /**
     * @type {TableHelper|exports|module.exports}
     */
    this.objTable = new TableHelper(tableName, this.dbKey);
    /**
     * @type {Redis}
     */
    this.objRedis = oujRedis.init(this.cacheKey);

    return this;
}

function getRowKey(args) {
    var key = this.cacheInfo['key'];
    var keys = key.split(',');
    var cacheKeys = [];
    for (var key in keys) {
        key = key.trim();
        if (args[key]) {
            if (util.isArray(args[key])) {
                cacheKeys.push(key + "=" + join('|', args[key]));
            } else {
                cacheKeys.push(key + "=" + args[key]);
            }
        }
    }

    var cacheKey = this.tableName + ":row";
    if (cacheKey) {
        return cacheKey + ':' + cacheKeys.join(':');
    } else {
        return cacheKey;
    }
}

function getAllKey(args) {
    var cacheKey = this.tableName + ":all";
    var otherCacheKey = cacheKey + ":others";

    var key = this.cacheInfo['all_key'];
    if (key) {
        var keys = explode(',', key);
        var cacheKeys = [];
        for (var key in keys) {
            key = key.trim();
            if (util.isArray(args[key]) || typeof args[key] == 'undefined' || args[key] === null) {
                // 如果有数组，或者有key不存在，则归类到others
                return otherCacheKey;
            } else {
                cacheKeys.push(key + "=" + args[key]);
            }
        }

        if (cacheKeys) {
            return cacheKey + ':' + cacheKeys.join(':');
        } else {
            return otherCacheKey;
        }
    } else {
        return otherCacheKey;
    }
}

function setCache(args) {
    var cacheKey = getRowKey(args);
    if (!cacheKey) {
        var msg = "没设置key:cacheKey," . JSON.stringify(this.cacheInfo);
        throw new Error(msg, CODE_REDIS_ERROR);
    }

    var pipeline = this.objRedis.pipeline();
    pipeline.hmset(cacheKey, args);
    if (this.cacheInfo['ttl'] > 0) {
        this.objRedis.expire(cacheKey, this.cacheInfo['ttl']);
    }

    return pipeline.exec();
}

function updateCache(args, cb) {
    var cacheKey = this.getRowKey(args);
    if (!cacheKey) {
        var msg = "没设置key:cacheKey," . json_encode(this.cacheInfo);
        throw new RedisException(msg, CODE_REDIS_ERROR);
    }

    var deferred = Q.defer();
    var ret = this.objRedis.exists(cacheKey);
    ret.then(function(result) {
        if (result == 1) {
            var pipeline = this.objRedis.pipeline();
            pipeline.hmset(cacheKey, args);

            if (this.cacheInfo['ttl'] > 0) {
                pipeline.expire(cacheKey, this.cacheInfo['ttl']);
            }

            var result = pipeline.exec();
            result.then(deferred.resolve, deferred.reject);
        }
    }, deferred.reject);

    return deferred.promise.nodeify(cb);
}

/**
 * 增加一行数据
 * @param args
 * @param boolean updateList
 * @return int
 */
Redis2MySql.prototype.addObject = function addObject(args, updateList) {
    var flag = this.objTable.addObject(args);
    updateList = is_bool(updateList) ? updateList : true;

    if (updateList) {
        this.delListCache(args);
    }

    return flag;
}

/**
 * 获取一个key的数据
 * @param where
 * @return <NULL, array>
 */
Redis2MySql.prototype.getRow = function getRow(where, cb) {
    var cacheKey = getRowKey(where);
    var deferred = Q.defer();

    var ret = this.objRedis.hgetall(cacheKey);
    ret.then(function(data) {
        if (!data) {
            // 从数据库重建
            var ret2 = this.objTable.getRow(where);
            ret2.then(function(row) {
                setCache(row);
                deferred.resolve(row);
            }, deferred.reject);
        } else {
            deferred.resolve(data);
        }
    }, deferred.reject);

    return deferred.promise.nodeify(cb);
}


/**
 * 读取多行数据
 * @param where
 * @param keyWord 查询关键字, array('_field', '_where', '_limit', '_sortKey', '_sortDir', '_lockRow', '_tableName')
 * @param  bool updateList 是否强制更新缓存
 * @return array:
 */
Redis2MySql.prototype.getAll = function(where, keyWord, updateList, cb) {
    var args = array_merge(where, keyWord);
    var key = http_build_query(args);
    if (key.length > 32) {
        key = md5(key);
    } else {
        key = http_build_query(args);
    }

    var deferred = Q.defer();
    var cacheKey = getAllKey(where) + ":" + key;
    if (!updateList) {
        this.objRedis.get(cacheKey)
            .then(function(data) {
                if (data) {
                    data = JSON.parse(data);
                    deferred.resolve(data);
                } else {
                    getAllFromDb(deferred);
                }
            }, deferred.reject);
    } else {
        getAllFromDb(deferred);
    }

    return deferred.promise.nodeify(cb);
}

function getAllFromDb(deferred) {
    var data = this.objTable.getAll(where, keyWord);
    this.objRedis.set(cacheKey, JSON.stringify(data));

    if (this.cacheInfo['ttl'] > 0) {
        this.objRedis.expire(cacheKey, this.cacheInfo['ttl']);
    }

    return data;
}

/**
 * 删除行的缓存
 * @param where
 * @throws RedisException
 */
Redis2MySql.prototype.delRowCache = function(where) {
    cacheKey = this.getRowKey(where);
    if (!cacheKey) {
        msg = "没设置key:cacheKey," . json_encode(this.cacheInfo);
        throw new RedisException(msg, CODE_REDIS_ERROR);
    }

    this.objRedis.del(cacheKey);
}

/**
 * 删除列表的缓存
 * @param where
 */
Redis2MySql.prototype.delListCache = function(where = []) {
    cacheKey = this.getAllKey(where);
    otherKey = "{this.tableName}:all:others";
    if (cacheKey === otherKey) {
        // 需要清除所有key删除
        cacheKey = "{this.tableName}:all";
        keys = this.objRedis.keys(cacheKey . '*');
    } else {
        // 除了删除当前keys，还需要删除others
        keys = this.objRedis.keys(cacheKey . '*');
        keys1 = this.objRedis.keys(otherKey . '*');
        keys = array_merge(keys, keys1);
    }

    if (!empty(keys)) {
        this.objRedis.del(keys);
    }
}

/**
 * 设置一个key的数据
 * @param args
 * @param boolean updateList
 * @return int 影响行数
 */
Redis2MySql.prototype.replaceObject = function(args, updateList = true) {
    flag = this.objTable.replaceObject(args);
    this.setCache(args);

    if (updateList) {
        this.delListCache(args);
    }

    return flag;
}

/**
 * 修改一个key的数据
 * @param args 更新的内容
 * @param where 更新的条件
 * @param boolean updateList
 * @return int 影响行数
 */
Redis2MySql.prototype.updateObject = function(args, where, updateList = true) {
    flag = this.objTable.updateObject(args, where);
    args = array_merge(args, where);
    this.updateCache(args);

    if (updateList) {
        this.delListCache(args);
    }

    return flag;
}



module.exports = Redis2MySql;
